# this part is copied from Brady Johnston, which is used to find classes in all addon modules.
# --------------------------------------------------------------------------------------------
import bpy
import inspect
import pkgutil
import importlib
from pathlib import Path


def init():
    global modules
    global classes

    modules = get_all_submodules(Path(__file__).parent)
    classes = get_register(modules)
    return classes


# Import modules
#################################################

def get_all_submodules(directory):
    return list(iter_submodules(directory, directory.name))

def iter_submodules(path, package_name):
    for name in sorted(iter_submodule_names(path)):
        yield importlib.import_module("." + name, package_name)

def iter_submodule_names(path, root=""):
    for _, module_name, is_package in pkgutil.iter_modules([str(path)]):
        if is_package:
            sub_path = path / module_name
            sub_root = root + module_name + "."
            yield from iter_submodule_names(sub_path, sub_root)
        else:
            yield root + module_name


# Find classes to register
#################################################

def get_register(modules):
    my_classes = list(set(iter_my_classes(modules)))
    return my_classes

def iter_my_classes(modules):
    base_types = get_register_base_types()
    for cls in get_classes_in_modules(modules):
        if any(base in base_types for base in cls.__bases__):
            if not getattr(cls, "is_registered", False):
                yield cls

def get_classes_in_modules(modules):
    classes = set()
    for module in modules:
        for cls in iter_classes_in_module(module):
            classes.add(cls)
    return classes

def iter_classes_in_module(module):
    for value in module.__dict__.values():
        if inspect.isclass(value):
            yield value

def get_register_base_types():
    return set(getattr(bpy.types, name) for name in [
        "Panel", "Operator", "PropertyGroup",
        "AddonPreferences", "Header", "Menu",
        "Node", "NodeSocket", "NodeTree",
        "GeometryNodeCustomGroup","ShaderNodeCustomGroup",
        "UIList", "RenderEngine",
        "Gizmo", "GizmoGroup",
    ])